﻿local _KEY,_NAME,_VER = "JD","MultiPanel","0.2 alpha" --or StackPanels?
local minfarbuild = 3892 -- requires Far 3.0 build >=3892 (hDlg:send)

local F = far.Flags
local build = select(4,far.AdvControl(F.ACTL_GETFARMANAGERVERSION,true)) --3,0,0,3892,0
if build<minfarbuild
   and 2~=far.Message(("FAR 3 build %s or newer required.\nYour build: %s")
                      :format(minfarbuild,build),_NAME,"Cancel;Continue","lw")
     then return
end

------------------------------------------------
--Settings

local macroKey = "`"         -- Ctrl~ / CtrlShift~

-------------------------------------------------------------
--- some general definitions
local function LockState(xLock,State) --set State: 0|1|2=toggle
                                      --get State: nil|-1
  local Lock = {Num=0, Caps=1, Scroll=2}
  local modeIsOn,keyIsPressed = 1,0xff80
  if State then win.Sleep(80) else State=-1 end
  local r = mf.flock(assert(Lock[xLock]),State)
  return band(r,modeIsOn)==modeIsOn,
         band(r,keyIsPressed)==keyIsPressed
end

--[[
RAlt,LAlt  = 0x1,0x2
RCtrl,LCtrl = 0x4,0x8
Shift = 0x10
NumLock,ScrollLock,CapsLock = 0x20,0x40,0x80
ENHANCED_KEY = 0x100
--]]
local RCtrl,LCtrl = 0x4,0x8
local anyCtrl = bor(RCtrl,LCtrl)
local function ModState(Mod) return band(Mouse.LastCtrlState,Mod) end -- State after last mf.waitkey!!

local function MessagePopup(msg,title,flags,delay)
  local s = far.SaveScreen()
  far.Message(msg,title or "","",flags)
  win.Sleep(delay or 500); far.RestoreScreen(s)
end

local function eq(a,b) --simple compare tables
  for k,v in pairs(a) do if b[k]~=v then return end end
  return true
end

--------------------------------------------------------
-- MultiPanel panel api logic
local ACTIVE,PASSIVE = F.PANEL_ACTIVE,F.PANEL_PASSIVE
local function getSelection(p,Info)
  if Info.SelectedItemsNumber==1 then
    local Item = panel.GetSelectedPanelItem(p,nil,1)
    return band(Item.Flags,F.PPIF_SELECTED)~=0 and {Item.FileName}
  elseif Info.SelectedItemsNumber>0 then
    local Sel = {}
    for i=1,Info.SelectedItemsNumber do
      Sel[i] = panel.GetSelectedPanelItem(p,nil,i).FileName
    end
    return Sel
  end
end

local function setSelection(p,Sel)
  if not Sel then return end
  local Info = panel.GetPanelInfo(p)
  local idx = {}
  for i=1,Info.ItemsNumber do
    local FileName = panel.GetPanelItem(p,nil,i).FileName
    for j=1,#Sel do
      if Sel[j]==FileName then idx[#idx+1]=i end
    end
  end
  panel.SetSelection(p,nil,idx,true)
end

local function getPanel(p)
  local Info = panel.GetPanelInfo(p)
  local info = {}
  if band(Info.Flags,F.PFLAGS_PLUGIN)==F.PFLAGS_PLUGIN then
    info = {
      isPlugin=true,
      --todo: exclusion for Temporary panel etc
      notShortcut=band(Info.Flags,F.PFLAGS_SHORTCUT)~=F.PFLAGS_SHORTCUT,
      Format=panel.GetPanelFormat(p),
      Prefix=panel.GetPanelPrefix(p),
    }
  end
  info.Info=Info
  info.Sel=getSelection(p,Info)
  info.Directory=panel.GetPanelDirectory(p)
  return info
end

local function setPanel(p,info)
  if not info then return end --debug
  if not panel.SetPanelDirectory(p,nil,info.Directory) then MessagePopup"oops"return end
  --http://bugs.farmanager.com/view.php?id=2735
  if not eq(panel.GetPanelDirectory(p),info.Directory) then return end
  local Info = info.Info
  local Flags = Info.Flags
  panel.SetSortMode(p,nil,Info.SortMode)
  panel.SetNumericSort(p,nil,band(Flags,F.PFLAGS_NUMERICSORT)~=0)
  panel.SetSortOrder(p,nil,band(Flags,F.PFLAGS_REVERSESORTORDER)~=0)
  panel.SetViewMode(p,nil,Info.ViewMode)
  setSelection(p,info.Sel)
  panel.RedrawPanel(p,nil,info.Info)
  return true
end

local DefPanel = {
  Directory = {
    Name=win.GetEnv"SystemDrive".."\\",
    File="",Param="",PluginId=win.Uuid"00000000-0000-0000-0000-000000000000",
  },
  Info = {
    SortMode=F.SM_EXT,--http://api.farmanager.com/ru/defs/sortmetods.html
    Flags=F.PFLAGS_NONE,--bor(F.PFLAGS_NUMERICSORT,F.PFLAGS_REVERSESORTORDER),
    ViewMode=2,
  }
}

--------------------------------------------------------
-- MultiPanel stack logic
local Panels = {}
function Panels:Push(p,info)
  info = info or getPanel(p)
  if info.notShortcut then
    panel.ClosePanel(p)
    panel.RedrawPanel(p)
    info = getPanel(p)
  end
  table.insert(self,1,info)
  return info
end
function Panels:Clone(i,info)
  self:CheckBounds(i)
  info = info or self[i]
  --if info.notShortcut then return end --debug
  table.insert(self,i,info)
  return info
end
function Panels:Remove(i)
  self:CheckBounds(i)
  return table.remove(self,i)
end
function Panels:Close(p,info,force)
  if #self>0 and not force then
    return self:Pop(p,1)
  else
    info = info or getPanel(p)
    if info.isPlugin then
      panel.ClosePanel(p)
      panel.RedrawPanel(p)
    else
      setPanel(p,DefPanel)
    end
    return true
  end
end
function Panels:Pop(p,i)
  self:CheckBounds(i)
  if setPanel(p,self[i]) then
    self:Remove(i)
    return true
  end
end
function Panels:Swap(p,info,i)
  self:CheckBounds(i)
  info = info or getPanel(p)
  if self:Pop(p,i)
      or not eq(panel.GetPanelDirectory(p),info.Directory) then
    self:Push(p,info) 
  end
end
function Panels:CheckBounds(i) --debug
  return assert(i and i>0 and i<=#self,"out of bounds")
end
local mt = {__index=Panels}
local Panels

--load/save
local S -- Settings
local function delSettings()
  mf.mdelete(_KEY,_NAME)
end
local function saveSettings()
  S.Panels=Panels
  mf.msave(_KEY,_NAME,S)
end
local function loadSettings()
  S = mf.mload(_KEY,_NAME) or {}
  Panels = S.Panels or Panels or {}
  setmetatable(Panels,mt)
end
loadSettings()

Event { description="[MultiPanel] Save settings (on exit)";
  group="ExitFAR";
  action=saveSettings;
}

local MPid = "D3656A7D-A75F-40F1-AB9B-59D962CE2224"
local function isMultiPanelMenu() return Menu.Id==MPid end;

local closeOnRelease
Event{ description="[MultiPanel] (close on release)"; group="DialogEvent";
  condition=function(Event,Param)
    return closeOnRelease
       and Event==F.DE_DLGPROCINIT and Param.Msg==F.DN_ENTERIDLE
       and isMultiPanelMenu() and ModState(anyCtrl)==0
  end;
  action=function(Event,Param)
    mf.postmacro(Keys,"Enter")
    --mf.postmacro(Keys,"CtrlEsc")
  end;
}

------------------------------------------------
-- MultiPanel-menu macros
Macro { description="[MultiPanel] Prev/Next";
  area="Menu"; key="/^[LR]Ctrl(Shift)?(CapsLock|Tab|`)$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):find"Shift" and "Up" or "Down") end;
}

Macro { description="[MultiPanel] Prev/Next (additional)";
  area="Menu"; key="/^[LR]Ctrl(Up|Down|PgUp|PgDn|Num9|Num3)$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):match"^R?Ctrl(.+)$") end;
}

Macro { description="[MultiPanel] Scroll long titles";
  area="Menu"; key="CtrlRight CtrlLeft";
  condition=isMultiPanelMenu;
  action=function() Keys("Alt"..akey(1):match"Ctrl(.+)") end;
}

Macro { description="[MultiPanel] Goto [0-9A-Z]";
  area="Menu"; key="/^[RL]Ctrl\\w$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):sub(-1)) end;
}

local TypeW = 10
local function prepareMenuItems()
  local Items = {}
  for i=0,#Panels do
    local info = Panels[i] or getPanel(ACTIVE)
    local hk = i
    hk = hk<=9 and hk or hk<=35 and string.char(("A"):byte()+hk-10)
    hk = hk and "&"..hk.."." or ""
    local Dir = info.Directory
    local Type = not info.isPlugin and "Panel" or info.Prefix:match"^[^:]*"
    Type = Type..(" "):rep(TypeW-Type:len())
    local file = Dir.File:match"[^\\]+$"
    local Name = Dir.Name:match".+" or "[Root]"
    Name = info.isPlugin and ((file and file or info.Format)..":"..Name) 
                          or Name
    local path = not info.isPlugin and Dir.Name or Dir.File~="" and Dir.File
    local notfound = path and not win.GetFileAttr(path)
    local Mark = notfound and "!" or info.Sel and "*"
    text = ("%2s %-$s %s %s"):gsub("%$",TypeW)
                             :format(hk,Type,Mark or " ",Name) --Sel
    Items[i+1] = {
      info=info,
      text=text,
      checked = not info.isPlugin and "≡", -- =≡■
      grayed = info.notShortcut or notfound,
      --disable
      --hidden
    }
  end
  return Items
end
local Items
local function new_pos(pos,key)
  if key~="" then
    local d = 
              key:find("Up$") and 1 or
              key:find("Down$") and -1 or
              key:find("^R?CtrlShift") and -1 or
              key:find("^R?Ctrl") and 1
    if d then
      pos = pos+d
      return pos>#Items and 1 or pos<1 and #Items or pos
    end
  end
end

local function expandBreakKeys(BreakKeysArr)
  local newBreakKeysArr = {}
  for _,item in ipairs(BreakKeysArr) do
    for key in item.BreakKey:gmatch("%S+") do
      local newitem = {}
      for k,v in pairs(item) do newitem[k]=v end
      newitem.BreakKey = key
      table.insert(newBreakKeysArr,newitem)
    end
  end
  return newBreakKeysArr
end
local BreakKeys = expandBreakKeys { --http://msdn.microsoft.com/library/dd375731
  {BreakKey="C+CONTROL",br=true},
  --{BreakKey="C+ESCAPE",br=true},
  {BreakKey="C+RETURN",  action=function(self,pos)
     if pos>1 then
       Panels:Swap(PASSIVE,nil,pos-1)
     else
       Keys"CtrlU Tab"
     end
   end,br = true},
  {BreakKey="RETURN",    action=function(self,pos)
     if pos>1 then
       Panels:Swap(ACTIVE,Items[1].info,pos-1)
     end
   end,br = true},
  {BreakKey="C+F3 F3",   action=function(self,pos)
     if pos>1 then
       Panels:Clone(pos-1,Items[pos].info)
     else
       Panels:Push(ACTIVE,Items[pos].info)
     end
     Items = prepareMenuItems()
   end},
  {BreakKey="C+F4 F4",   action=function(self,pos)
     if pos==1 then
       Panels:Close(ACTIVE,Items[1].info)
     else
       Panels:Remove(pos-1)
     end
     Items = prepareMenuItems()
   end},
  {BreakKey="CS+F9 S+F9",action=saveSettings},
  {BreakKey="C+F9 F9",   action=function()
     if far.Message("Reload settings?",_NAME,";OkCancel") then
       loadSettings()
       Items = prepareMenuItems()
     end
   end},
  {BreakKey="C+F8 F8",   action=function()
     if far.Message("Delete saved settings?",_NAME,";OkCancel") then
       delSettings()
       Panels = nil
       loadSettings()
       Items = prepareMenuItems()
     end
   end},
  {BreakKey="C+F1 F1",   action=function(self)
     far.Message([[

See Bottom title for hotkeys

[to be continued...]

]],_NAME.." r".._VER,nil,"kl")
     self.br = closeOnRelease and ModState(anyCtrl)==0
   end},
}
local Props = {Title="MultiPanel",Id=win.Uuid(MPid),Flags=F.FMENU_WRAPMODE}
local Bottom = "F3:Clone, F4:Close, Release:Swap, Ctrl-Enter:Swap with passive"

local function showMultiPanel(key)
  Items = prepareMenuItems()
  local item = Items[1]
  local pos = item.grayed and 1 or new_pos(1,key)
  if not (key=="donotcloseOnRelease" or ModState(anyCtrl)==0) then closeOnRelease = true end
  repeat
    Props.Bottom = Bottom
    Props.SelectIndex = pos
    item,pos = far.Menu(Props, Items, BreakKeys)
    if not item then
      break
    elseif item.action then --BreakKey
      item:action(pos)
    else assert(item.br) --debug
    end
  until item.br
  closeOnRelease = false
  return pos
end

--todo appearance
--todo Help
--todo Info QView Tree ShellAutoCompletion
--todo PFLAGS_VISIBLE
--todo Extra Settings
--todo local Settings?
--todo autoswitch?

--local allPanels  = "Shell Info QView Tree Search ShellAutoCompletion"
local allPanels  = "Shell Search"
Macro { description="[MultiPanel] switch";
  area=allPanels; key="/^LCtrl(Shift)?("..macroKey..")$/";
  action=function()
    showMultiPanel(akey(1))
    if macroKey=="CapsLock" then LockState("Caps",0) end
  end;
}
Macro { description="[MultiPanel] switch";
  area=allPanels; key="LCtrl";
  action=function()
    showMultiPanel"donotcloseOnRelease"
  end;
}
Macro { description="[MultiPanel] goto Panels";
  area="Editor Viewer"; key="/^LCtrl(Shift)?("..macroKey..")$/";
  action=function()
    Keys"F12 0"
    if Area.Shell then showMultiPanel "" end
    if macroKey=="CapsLock" then LockState("Caps",0) end
  end;
}

-----------------
-- HoldMacro
local HoldDelay = 200
local function HoldMacro(t)
  local action = t.action
  t.action=function()
    far.Timer(10,function(timer,start,key,area)
      local released = ModState(anyCtrl)==0
      if Far.UpTime-start<HoldDelay and not released then return end
      timer:Close()
      mf.postmacro(function()
        if area==Area.Current then
          if released then
            Keys(key)
          else
            --??check condition
            action()
          end
        end
      end)
    end,Far.UpTime,akey(1),Area.Current)
  end
  return Macro(t)
end
HoldMacro { description="[MultiPanel] store active panel";
  area=allPanels; key="LCtrlF3";
  action=function()
    local p = ACTIVE
    local info = getPanel(p)
    if info.notShortcut then
      mf.postmacro(MessagePopup,("Unable to store this plugin panel: %s"):format(info.Format),_NAME,"w",delay)
      showMultiPanel"donotcloseOnRelease"
    else
      Panels:Push(p,info)
      MessagePopup("Storing panel...",_NAME,flags,100)
    end
  end;
}
HoldMacro { description="[MultiPanel] store passive panel";
  area=allPanels; key="LCtrlShiftF3";
  action=function()
    local p = PASSIVE
    local info = getPanel(p)
    if info.notShortcut then
      mf.beep()
      MessagePopup(("Unable to store this plugin panel: %s"):format(info.Format),_NAME,"w",delay)
    else
      Panels:Push(p,info)
      MessagePopup("Storing panel...",_NAME,flags,100)
    end
  end;
}
HoldMacro { description="[MultiPanel] restore prev. panel";
  area=allPanels; key="LCtrlF4";
  action=function()
    Panels:Close(ACTIVE)
    MessagePopup("Closing panel...",_NAME,flags,100)
  end;
}
HoldMacro { description="[MultiPanel] restore prev. panel as passive";
  area=allPanels; key="LCtrlShiftF4";
  action=function()
    Panels:Close(PASSIVE)
    MessagePopup("Closing panel...",_NAME,flags,100)
  end;
}
HoldMacro { description="[MultiPanel] close active panel";
  area=allPanels; key="LCtrlF8";
  action=function()
    Panels:Close(ACTIVE,nil,"nopop")
  end;
}
HoldMacro { description="[MultiPanel] close passive panel";
  area=allPanels; key="LCtrlShiftF8";
  action=function()
    Panels:Close(PASSIVE,nil,"nopop")
  end;
}
